(function () {
    var componentClass = 'paginator-component',
        unpaginatedClass = 'paginator-unpaginated',
        modelClass = 'model',
        viewClass = 'view',
        watchClass = 'watch',
        pageClass = 'page',
        pageBlockClass = 'page-block',
        pageAddedClass = 'page-added',
        pageDeletedClass = 'page-deleted',
        mainClass = 'content',
        headerClass = 'header',
        footerClass = 'footer',
        marginClass = 'margin',
        parentDataAttrName = 'parent',
        terminalClass = '-terminal',
        blockParentClass = 'block-parent',
        orderDataAttrName = 'data-order',
        idAttrName = 'data-element-id',
        controlElsSelector = 'input, select, textarea, button, a, meter',
        focusableElsSelector = '[tabindex="0"]',
        debounceDuration = 250;

    // http://stackoverflow.com/questions/7051897/how-can-i-select-the-shallowest-matching-descendant
    jQuery.fn.findShallow = function(sel) {
        return (function findShallow(root, sel) {
            var children = root.children();
            if (children.length) {
                var matching = children.filter(sel);
                if (matching.length) {
                    return matching.first();
                }
                return findShallow(children, sel);
            }
            return $();
        })(this, sel);
    };

    // http://james.padolsey.com/snippets/sorting-elements-with-jquery/
    jQuery.fn.sortElements = (function () {
        var sort = [].sort;

        return function (comparator, getSortable) {

            getSortable = getSortable || function () { return this; };

            var placements = this.map(function () {

                var sortElement = getSortable.call(this),
                    parentNode = sortElement.parentNode,

                // Since the element itself will change position, we have
                // to have some way of storing its original position in
                // the DOM. The easiest way is to have a 'flag' node:
                    nextSibling = parentNode.insertBefore(
                        document.createTextNode(''),
                        sortElement.nextSibling
                    );

                return function () {

                    if (parentNode === this) {
                        throw new Error(
                            "You can't sort elements if any one is a descendant of another."
                        );
                    }

                    // Insert before flag:
                    parentNode.insertBefore(this, nextSibling);
                    // Remove flag:
                    parentNode.removeChild(nextSibling);

                };

            });

            return sort.call(this, comparator).each(function (i) {
                placements[i].call(getSortable.call(this));
            });

        };
    })();

    jQuery.fn.reverse = [].reverse;

    /**
     * Creates a debounced version of a function
     * @param {function} fn The function
     * @param {number} duration The debounce duration.
     * @returns {function} The debounced version of the function.
     */
    function debounce(fn, duration) {
        var debounceHandle = null;

        return function () {
            var args = Array.prototype.slice.call(arguments);

            if (!!debounceHandle) {
                clearTimeout(debounceHandle);
            }

            debounceHandle = setTimeout(function () {
                fn.apply(null, args);
            }, duration);
        };
    }

    /**
     * Converts a data attribute key/value pair to its corresponding CSS selector.
     * @param key The attribute key.
     * @param value The attribute value.
     * @returns {string} The selector string.
     */
    function toAttrSelector(key, value) {
        return '[' + key + '="' + value + '"]';
    }

    /**
     * Converts a class string to a CSS selector.
     * @param {string} classString The class string.
     * @returns {string} The selector string.
     */
    function toClassSelector(classString) {
        return '.' + classString.split(' ').filter(function (string) { return string.trim().length > 0 }).join('.');
    }

    /**
     * Class for the options for the paginator.
     * @param {object} rawOpts The raw options data.
     * @constructor
     */
    function PaginatorOptions(rawOpts) {
        rawOpts = rawOpts || {};

        return {
            watch: !!rawOpts.watch
        };
    }

    /**
     * Class for the container of contents on a page.
     * @param {string} klass The class name of the main HTML element generated by the component.
     * @param {PaginatorOptions} opts The options of the paginator.
     * @constructor
     */
    function PaginatorContainer(klass, opts) {
        var self = this,
            focusedEl = null,
            focused,
            page;

        /**
         * Creates the elements of the component.
         */
        function createElement() {
            self.$el = self.$el || $('<div>');
            self.$margin = self.$margin = $('<div>').addClass(marginClass);

            if (!self.$el.hasClass(klass)) {
                self.$el.addClass(klass);
                self.$el.on('focus', 'input, select, textarea, option, button, a, [tabindex]', function () {
                    focusedEl = this;
                    focused = true;
                });
            }

            self.$el.append(self.$margin);
        }

        self.retainFocus = function retainFocus() {
            if (focusedEl === null) {
                return;
            }
            self.$el.find(toAttrSelector('data-element-id', $(focusedEl).data('elementId'))).focus();
        };

        self.isFocused = function isFocused() {
            return focused;
        };

        /**
         * Appends a content to this component.
         * @param {jQuery} $el The element to append.
         */
        self.append = function append($el) {
            self.$margin.append($el);
        };

        /**
         * Clears the page content.
         */
        self.clear = function clear() {
            self.$margin.html('');
        };

        /**
         * Mounts this component to a parent component.
         * @param {object} parent The parent component.
         */
        self.mountTo = function mountTo(parent) {
            if (page === parent) {
                return;
            }

            page = parent;
            page.$el.append(self.$el);
        };

        /**
         * Gets the content blocks of this component.
         * @returns {jQuery} The content blocks of this component.
         */
        self.getBlocks = function getBlocks() {
            return self.$margin.children();
        };

        createElement();
    }

    /**
     * Class for the page that is to be rendered in the paginator.
     * @param {PaginatorOptions} opts The options of the paginator.
     * @constructor
     */
    function PaginatorPage(opts) {
        var self = this,
            paginator;

        self.header = new PaginatorContainer(headerClass, opts);
        self.content = new PaginatorContainer(mainClass, opts);
        self.footer = new PaginatorContainer(footerClass, opts);

        /**
         * Creates the elements of the component.
         */
        function createElement() {
            self.$el = self.$el || $('<div>').addClass(pageClass);
            self.$el
                .append(self.header.$el)
                .append(self.content.$el)
                .append(self.footer.$el);
        }

        /**
         * Gets the current page index.
         * @returns {number} The current page index.
         */
        function getPageIndex() {
            var pageIndex = -1;

            paginator.pages.forEach(function (page, i) {
                if (page !== self) {
                    return;
                }

                pageIndex = i;
            });

            return pageIndex;
        }

        /**
         * Gets the lower boundary of the page.
         * @returns {number} The lower boundary of the page.
         */
        function getContentLowerBoundary() {
            var $footer = self.footer.$el,
                $margin = self.content.$margin,
                bottomHeight = Math.max($footer.length > 0 ? $footer.height() : 0, parseInt($margin.css('margin-bottom')));
                //topHeight = Math.max($header.length > 0 ? $header.height() : 0, parseInt($margin.css('margin-top')));

            //var $footer = self.footer.$el,
            //    $margin = self.content.$margin;

            //return self.$el.height();
            //return self.$el.height() - ($footer.length > 0 ? $footer.height() : parseInt($margin.css('margin-bottom')));
            //return topHeight + getContentHeight();

            return self.$el.height() - bottomHeight;
        }

        /**
         * Determines if the page has elements that can trigger page breaks.
         * @returns {boolean} Does the page require breaks?
         */
        self.hasBreaks = function hasBreaks() {
            return self.getOverflowBlocks().length > 0;
        };

        /**
         * Determines if the page is blank (no blocks currently written).
         * @returns {boolean} Is the page blank?
         */
        self.isBlank = function isBlank() {
            return self.content.$margin.children().length < 1;
        };

        /**
         * Gets the elements that overflow the content area of the page.
         * @returns {jQuery} The elements.
         */
        self.getOverflowBlocks = function getOverflowBlocks() {
            return self.content.getBlocks()
                .filter(function () {
                    var $block = $(this),
                        position = $block.position().top,
                        height = $block.height(),
                        isOverflow = position + height > getContentLowerBoundary();

                    return isOverflow;
                });
        };

        /**
         * Unmounts the page.
         */
        self.unmount = function unmount() {
            self.$el.remove();
        };

        /**
         * Mounts the component to a parent.
         * @param {object} parent The component
         */
        self.mountTo = function mountTo(parent) {
            if (paginator === parent) {
                return;
            }

            paginator = parent;
            paginator.view.$el.append(self.$el);
        };

        createElement();
    }

    /**
     * Class for the renderer of the paginator.
     * @param {PaginatorOptions} opts
     * @constructor
     */
    function PaginatorRenderer(opts) {
        var self = this,
            model,
            view,
            paginator,
            firstPage = new PaginatorPage(opts),
            isRendering = false,
            lastFocusedElementId,
            $lastFocusedEl = [];

        /**
         * Writes the appropriate headers and the footers of the page.
         */
        function writePageComponents() {
            paginator.pages.forEach(function (page, i) {
                writeToBlockContainer(headerClass, page, i);
                writeToBlockContainer(footerClass, page, i);
            });
        }

        /**
         * Determines if the content class is repetitive.
         * @param klass The content class.
         * @returns {boolean} Is the content class repetitive among pages?
         */
        function isContentClassRepetitive(klass) {
            return klass === headerClass || klass === footerClass;
        }

        /**
         * Gets a page's content container of a content class.
         * @param {PaginatorPage} page The page.
         * @param {string} klass The content class.
         * @returns {PaginatorContainer} The content container.
         */
        function getContentContainer(page, klass) {
            if (klass === mainClass) {
                return firstPage.content;
            }

            return page[klass];
        }

        /**
         * Gets the CSS key of the page to allocate space for the content class.
         * @param klass The content class.
         * @returns {string|null} The CSS key.
         */
        function getAllocationKey(klass) {
            switch (klass) {
                case headerClass:
                    return 'margin-top';
                case footerClass:
                    return 'margin-bottom';
                default:
                    break;
            }

            return null;
        }

        /**
         * Gets the allocation size of a page's content class.
         * @param {PaginatorPage} page The page.
         * @param {string} klass The content class.
         * @returns {number} The margin value, in pixels.
         */
        function getAllocationValue(page, klass) {
            if (klass === headerClass || klass === footerClass) {
                return getContentContainer(page, klass).$el.height();
            }
            return 0;
        }

        /**
         * Allocate page space for the content of the content class.
         * @param {PaginatorPage} page The page.
         * @param {string} klass The content class.
         */
        function allocateSpaceForContent(page, klass) {
            var key = getAllocationKey(klass),
                value;

            if (!key) {
                return;
            }

            value = getAllocationValue(page, klass);
            page.content.$margin.css(key, value);
        }

        /**
         * Reset default margins for the page.
         * @param {PaginatorPage} page
         * @param {string} klass
         */
        function resetToDefaultMargins(page, klass) {
            var key;

            if (!page) {
                return;
            }

            key = getAllocationKey(klass);

            page.content.$margin.css(key, null);
        }

        /**
         *
         */
        function resetMarginsOfAllPages() {
            paginator.pages.forEach(function (page) {
                resetToDefaultMargins(page, headerClass);
                resetToDefaultMargins(page, footerClass);
            });
        }

        /**
         * Labels the control elements of the block, to be used for synchronizing
         * the data and attributes of the controls.
         * @param {Element} $block The block.
         * @param {number} index The index of the block.
         */
        function labelControlElements($block, index) {
            $block
                .find(controlElsSelector + ', ' + focusableElsSelector)
                .each(function (j) {
                    $(this).attr(idAttrName, '' + index + '.' + j);
                });
        }

        function isBlockDeleted($block) {
            return !$block.parents().is('body') ||
                $block.parents(toClassSelector(pageDeletedClass)).length > 0;
        }

        /**
         * Writes the content of the page.
         */
        function writeToBlockContainer(klass, page, i) {
            var headerFooterIndex = model.getIndexForPage(klass, i),
                blocks,
                isRepetitive = isContentClassRepetitive(klass),
                $contentBlock,
                contentContainer;

            contentContainer = getContentContainer(page, klass);

            if (!contentContainer) {
                // No content container to put the content blocks to
                return;
            }

            if (isRepetitive) {
                // Clear repetitive content because
                // we can clone them again.
                contentContainer.clear();
            }

            if (headerFooterIndex < 0) {
                // No header/footer in this page.
                return;
            }

            blocks = model.getBlocks(klass);

            if (isRepetitive) {
                // get appropriate header/footer index for this page
                blocks = blocks(headerFooterIndex);
            }

            blocks
                .each(function (i) {
                    var $block = $(this),
                        $blockParent = $block.parent(),
                        $renderedBlock = $block;

                    labelControlElements($block, i);

                    if (isRepetitive) {
                        $renderedBlock = $block.clone(true, true);
                    }

                    if (isBlockDeleted($block)) {
                        // block is detached and not visible, or is deleted.
                        $renderedBlock.off('change', controlElsSelector);
                        return;
                    }

                    $blockParent = $block.parent();

                    $blockParent
                        .addClass(blockParentClass)
                        .attr(orderDataAttrName, 0);

                    $renderedBlock
                        .attr(idAttrName, i)
                        .data(parentDataAttrName, $blockParent)
                        .removeClass(pageBlockClass)
                        .addClass(pageAddedClass)
                        .on('change', controlElsSelector, function (e, d) {
                            var $origEl,
                                $this = $(this);

                            if (!isRepetitive) {
                                return;
                            }

                            lastFocusedElementId = $this.attr(idAttrName);
                            $origEl = $blockParent.find(toAttrSelector(idAttrName, lastFocusedElementId));

                            if ($this.is('select')) {
                                $origEl
                                    .find('option')
                                    .removeAttr('selected')
                                    .filter(toAttrSelector('value', $this.val()))
                                    .attr('selected', 'selected');
                            }

                            $origEl
                                .val($this.val())
                                .trigger(e.type, d);
                        });

                    contentContainer.append($renderedBlock);
                });

            contentContainer.getBlocks().length > 0 ? allocateSpaceForContent(page, klass) : resetToDefaultMargins(page, klass);
            retainFocus();
        }

        function getBlockContainerOfBlock(block) {
            return $(block).parent(
                toClassSelector(headerClass) + ',' +
                toClassSelector(footerClass) + ',' +
                toClassSelector(mainClass)
            );
        }

        function retainFocus() {
            paginator.pages.forEach(function (page) {
                if (page.header.isFocused()) {
                    page.header.retainFocus();
                    return;
                }
                if (page.footer.isFocused()) {
                    page.footer.retainFocus();
                    return;
                }
                if (page.content.isFocused()) {
                    page.content.retainFocus();
                    return;
                }
            });
            //var $blockContainer, blockContainerClass;
            //
            //lastFocusedElementId = $(document.activeElement).attr(idAttrName) || lastFocusedElementId;
            //$blockContainer = getBlockContainerOfBlock(document.activeElement);
            //
            //if ($blockContainer.length > 0) {
            //    blockContainerClass = $blockContainer.attr('class');
            //}
            //
            //// contentblock should be replaced by contentContainer.$el
            //$lastFocusedEl = $blockContainer.find(toAttrSelector(idAttrName, lastFocusedElementId));
            //
            //if ($lastFocusedEl.length > 0) {
            //    $lastFocusedEl[0].focus();
            //}


        }

        /**
         * Sets the ordering of each block rendered in the view.
         */
        function setOrder() {
            model.$el
                .find(toClassSelector(blockParentClass))
                .each(function (i) {
                    $(this).attr(orderDataAttrName, i);
                });

            view.getBlocks('content')
                .each(function () {
                    var $block = $(this),
                        $blockParent = $block.data(parentDataAttrName);

                    $block.attr(orderDataAttrName, $blockParent.attr(orderDataAttrName));
                });
        }

        /**
         * Checks the view for deleted blocks and performs the appropriate actions.
         * @returns {number} The index of the first page with blocks deleted from it.
         */
        function permanentlyDeleteBlocks() {
            var firstPageWithDeletedBlocks = null;

            view
                .getBlocks('content')
                .reverse()
                .each(function () {
                    var $block = $(this),
                        $blockParent,
                        isParentDeleted,
                        page;

                    $blockParent = $block.data(parentDataAttrName);

                    isParentDeleted = !$blockParent ||
                        $blockParent.hasClass(pageDeletedClass) ||
                        $blockParent.parents(toClassSelector(pageDeletedClass)).length > 0 ||
                        !$blockParent.parents().is('body');

                    if (!isParentDeleted) {
                        return;
                    }

                    page = $block.parents('.page').index();

                    if (firstPageWithDeletedBlocks === null || firstPageWithDeletedBlocks > page) {
                        firstPageWithDeletedBlocks = page;
                    }

                    $block.remove();
                });

            return firstPageWithDeletedBlocks;
        }

        /**
         * Checks the view for deleted blocks and performs the appropriate actions.
         */
        function checkDeletedBlocks() {
            var firstPageWithDeletedBlocks = permanentlyDeleteBlocks(),
                thePage;

            if (firstPageWithDeletedBlocks === null) {
                return;
            }

            paginator.pages.forEach(function (page, i) {
                if (i <= firstPageWithDeletedBlocks) {
                    thePage = page;
                    return;
                }
                if (!thePage) {
                    return;
                }
                page.content.getBlocks().each(function () {
                    thePage.content.append(this);
                });
            });
        }

        /**
         * Inserts a page for new content.
         * @param {{blocks:jQuery,pageIndex:number}} content The content.
         */
        function insertPageForContent(content) {
            var newPage = new PaginatorPage(opts);

            content.blocks.each(function () {
                newPage.content.append($(this));
            });

            paginator.pages.slice(content.pageIndex + 1).forEach(function (page) {
                newPage.content.append(page.content.getBlocks());
            });

            view.addPage(newPage);
        }

        /**
         * Gets the content of the first page that has a break.
         * @returns {{blocks:jQuery,pageIndex:number}} The content to insert.
         */
        function getFirstPageWithBreak() {
            var pageToBreak = null;

            paginator.pages.forEach(function (page, i) {
                if (!page.hasBreaks() || pageToBreak !== null) {
                    return;
                }

                // Get the first page to have page breaks.
                pageToBreak = {
                    pageIndex: i,
                    blocks: page.getOverflowBlocks()
                };
            });

            return pageToBreak;
        }

        /**
         * Performs page breaks among the view's pages.
         * @returns {boolean} A value that determines if the renderer performed page breaks, which is used to repeat this method.
         */
        function performPageBreaks() {
            var pageToBreak = getFirstPageWithBreak();

            if (pageToBreak === null) {
                return false;
            }

            insertPageForContent(pageToBreak);
            return true;
        }

        /**
         *
         */
        function removeBlankPages() {
            var pagesToDelete = [];

            paginator.pages.forEach(function (page, i) {
                if (!page.isBlank()) {
                    return;
                }
                pagesToDelete.push(i);
            });

            pagesToDelete
                .reverse()
                .forEach(function (pageNumber) {
                    if (paginator.pages.length < 2) {
                        return;
                    }
                    paginator.pages[pageNumber].unmount();
                    paginator.pages.splice(pageNumber, 1);
                });
        }

        /**
         *
         */
        function orderContent() {
            view
                .$el
                .find('.content')
                .find('.margin')
                .children()
                .sortElements(function (block1, block2) {
                    var order1 = parseInt($(block1).attr(orderDataAttrName)),
                        order2 = parseInt($(block2).attr(orderDataAttrName));

                    return order1 - order2;
                });
        }

        /**
         * Adds the first page if it does not exist.
         */
        function keepFirstPage() {
            if (paginator.pages.length > 0) {
                return;
            }
            view.addPage(firstPage);
        }

        /**
         * Writes the contents to the pages.
         */
        function doWriting() {
            var hasPerformedPageBreaks;
            do {
                hasPerformedPageBreaks = performPageBreaks();
                writePageComponents();
            } while (hasPerformedPageBreaks);
        }

        /**
         * Simulates a lock so that the renderer will not run multiple times.
         * @param {function} fn The function to run.
         */
        function lock(fn) {
            if (isRendering) {
                return;
            }

            paginator.detachObserver();
            isRendering = true;

            fn();

            paginator.attachObserver();
            isRendering = false;
        }

        /**
         * Renders the content.
         */
        self.render = function render() {
            function doRender() {
                //console.dir(document.activeElement); // start
                keepFirstPage();
                checkDeletedBlocks();
                // TODO implement hard page/section breaks
                // section breaks are what reset the header indexing, etc.
                // TODO move elements when extra space has been found on previous pages.
                resetMarginsOfAllPages();
                writeToBlockContainer(mainClass);
                setOrder();
                orderContent();
                doWriting();
                //console.dir(document.activeElement); // first write
                removeBlankPages();
                doWriting();
                retainFocus();
                //console.dir(document.activeElement); // second write
            }

            lock(function () {
                paginator.$el.trigger('paginator.renderstart');
                doRender();

                //if (paginator.$el.find('.content').find('img').length < 1) {
                //    setTimeout(function () {
                //        doRender();
                //        paginator.$el.trigger('paginator.renderend');
                //    }, 500);
                //    return;
                //}
                paginator.$el.imagesLoaded().always(function () {
                    doRender();
                    paginator.$el.trigger('paginator.renderend');
                });
            });
        };

        /**
         * Mounts the renderer to the paginator.
         * @param {Paginator} parent The paginator.
         */
        self.mountTo = function mountTo(parent) {
            paginator = parent;
            model = paginator.model;
            view = paginator.view;
        };
    }

    /**
     * Class for the model of the paginator.
     * @param {PaginatorOptions} opts The options of the paginator.
     * @constructor
     */
    function PaginatorModel(opts) {
        var self = this,
            paginator;

        /**
         * Creates the model element.
         */
        function createElement() {
            self.$el = self.$el || $('<div>').addClass(modelClass);
            if (!!self.$watch) {
                return;
            }
            self.$watch = self.$watch || $('<div>').addClass(watchClass);
            self.$el.append(self.$watch);
        }

        /**
         *
         * @param klass
         * @param isTerminal
         * @returns {*}
         */
        function getBlockContainers(klass, isTerminal) {
            return self.$watch.find(toClassSelector(!!isTerminal ? klass + ' ' + terminalClass : klass));
        }

        /**
         *
         * @param klass
         * @returns {*}
         */
        function getCount(klass) {
            return getBlockContainers(klass).length;
        }

        /**
         *
         * @param klass
         * @param pageNumber
         * @returns {number}
         */
        function getIndexForTerminalBlockContainer(klass, pageNumber) {
            var i,
                count = getCount(klass),
                index = -1,
                pageCount = paginator.view.getPageCount(),
                lastIndex = pageCount - 1,
                $blocks = getBlockContainers(klass);

            if (count === 1) {
                switch (klass) {
                    case headerClass:
                        return pageNumber === 0 ? 0 : -1;
                    case footerClass:
                        return pageNumber === lastIndex ? 0 : -1;
                    default:
                        break;
                }
            }

            i = 0;
            do {
                index = (index + 1) % count;
                if (pageNumber === 0 && klass === headerClass ||
                    pageNumber === lastIndex && klass === footerClass) {
                    $blocks.each(function (k) {
                        if (!$(this).hasClass(terminalClass)) {
                            return;
                        }
                        index = k;
                    });
                    return index;
                }

                if ($blocks.eq(index).hasClass(terminalClass)) {
                    index++;
                }

                i++;
            } while (i < pageNumber);

            return index;
        }

        /**
         *
         * @param klass
         * @returns {boolean}
         */
        self.hasTerminalBlockContainer = function hasTerminalBlockContainer(klass) {
            return getBlockContainers(klass, true).length > 0;
        };

        /**
         *
         * @param klass
         * @param pageNumber
         * @returns {*}
         */
        self.getIndexForPage = function getIndexForPage(klass, pageNumber) {
            if (!self.hasTerminalBlockContainer(klass)) {
                return pageNumber % getCount(klass);
            }

            return getIndexForTerminalBlockContainer(klass, pageNumber);
        };

        /**
         *
         * @param klass
         * @returns {*}
         */
        self.getBlocks = function getBlocks(klass) {
            switch (klass) {
                case headerClass:
                case footerClass:
                    return function (i) {
                        var $blockContainer = getBlockContainers(klass).eq(i),

                            isTerminal = $blockContainer.hasClass(terminalClass),
                            $blocks = $blockContainer.find(toClassSelector(pageBlockClass));

                        return isTerminal ?
                            $blocks.addClass(terminalClass) :
                            $blocks.removeClass(terminalClass);
                    };
                default:
                    break;
            }

            return getBlockContainers(klass).find(toClassSelector(pageBlockClass));
        };

        /**
         *
         * @param klass
         * @returns {jQuery}
         */
        self.getOrderedBlocks = function getOrderedBlocks(klass) {
            return getBlockContainers(klass).find('[' + orderDataAttrName + ']');
        };

        /**
         *
         * @param parent
         */
        self.mountTo = function mountTo(parent) {
            if (paginator === parent) {
                return;
            }

            paginator = parent;
            paginator.$el.append(self.$el);
        };

        createElement();
    }

    /**
     * Class for the view of the paginator.
     * @param {PaginatorOptions} opts The options of the paginator.
     * @constructor
     */
    function PaginatorView(opts) {
        var self = this;

        self.paginator = null;

        /**
         *
         */
        function createElement() {
            self.$el = self.$el || $('<div>').addClass(viewClass);
        }

        /**
         *
         * @param page
         * @param i
         */
        function addToPageList(page, i) {
            if (isNaN(i) || i === null) {
                self.paginator.pages.push(page);
                return;
            }

            self.paginator.pages.splice(i, 0, page);
        }

        /**
         *
         * @returns {Number}
         */
        self.getPageCount = function getPageCount() {
            return self.paginator.pages.length;
        };

        /**
         *
         * @param klass
         * @returns {jQuery}
         */
        self.getBlocks = function getBlocks(klass) {
            return self.$el
                .find(toClassSelector(klass))
                .children(toClassSelector(marginClass))
                .children();
        };

        /**
         *
         * @param page
         * @param i
         * @returns {PaginatorPage}
         */
        self.addPage = function addPage(page, i) {
            addToPageList(page, i);
            page.mountTo(self.paginator);
            return page;
        };

        /**
         *
         * @param parent
         */
        self.mountTo = function mountTo(parent) {
            self.paginator = parent;
            self.paginator.$el.append(self.$el);
        };

        createElement();
    }

    /**
     * Class for the paginator.
     * @param {Element} $el The element.
     * @param {object} opts The options of the paginator.
     * @constructor
     */
    function Paginator($el, opts) {
        var self = this,
            modelObserver = new MutationObserver(commitMutations),
            pub,
            areEventsAttached = false,
            isObserving = false,
            debouncedRender = debounce(render, debounceDuration);

        opts = new PaginatorOptions(opts);

        /**
         *
         * @type {PaginatorModel}
         */
        self.model = new PaginatorModel(opts);

        /**
         *
         * @type {PaginatorView}
         */
        self.view = new PaginatorView(opts);

        /**
         *
         * @type {PaginatorRenderer}
         */
        self.renderer = new PaginatorRenderer(opts);

        /**
         *
         * @type {Array}
         */
        self.pages = self.pages || [];

        /**
         *
         */
        function render() {
            self.renderer.render();
        }

        /**
         *
         * @param mutations
         */
        function commitMutations(mutations) {
            var firstMutation = mutations[0];

            if (firstMutation.target === self.model.$watch[0] &&
                firstMutation.type === 'attributes' &&
                firstMutation.attributeName === 'id') {
                return;
            }
            debouncedRender();
        }

        /**
         *
         * @param $el
         */
        function bindToElement($el) {
            if (!!self.$el) {
                self.$el.removeClass(componentClass);
            }
            self.$el = $el;
            if (self.$el.hasClass(componentClass)) {
                return;
            }
            self.$el
                .addClass(componentClass)
                .removeClass(unpaginatedClass);
        }

        /**
         *
         */
        function createModel() {
            var $modelChildren;

            self.model.mountTo(self);

            $modelChildren = self
                .$el
                .find(
                //.findShallow(
                    [
                        toClassSelector(headerClass),
                        toClassSelector(mainClass),
                        toClassSelector(footerClass)
                    ]
                        .join(',')
                );

            self.model.$watch.append($modelChildren);
        }

        /**
         *
         */
        function createView() {
            self.view.mountTo(self);
        }

        /**
         *
         */
        function createRenderer() {
            self.renderer.mountTo(self);
        }

        /**
         *
         */
        self.detachObserver = function detachObserver() {
            console.log('disconnected'); // don't remove this, dunno why
            //setTimeout(function () {
                modelObserver.disconnect();
            //}, 10);
        };

        /**
         *
         */
        self.attachObserver = function attachObserver() {
            modelObserver
                .observe(self.model.$watch[0], {
                    childList: true,
                    attributes: true,
                    characterData: true,
                    subtree: true
                });
        };

        /**
         *
         */
        self.attachEventListeners = function attachEventListeners() {
            self.model.$watch
                .on('change', controlElsSelector, function (e) {
                    if (!isObserving) {
                        return;
                    }

                    commitMutations([e]);
                });
        };

        /**
         *
         */
        function observeModel() {
            if (isObserving) {
                return;
            }
            if (opts.watch) {
                self.attachObserver();
            }
            isObserving = true;
        }

        /**
         *
         */
        function observeModelEvents() {
            if (areEventsAttached) {
                return;
            }
            self.attachEventListeners();
            areEventsAttached = true;
        }

        /**
         *
         */
        function doObserve() {
            observeModel();
            observeModelEvents();
        }

        /**
         *
         */
        function doRefresh() {
            render();
        }

        pub = (function PaginatorInstance() {
            return {
                observe: doObserve,
                refresh: doRefresh
            };
        })();

        bindToElement($el);
        createModel();
        createView();
        createRenderer();
        doObserve();
        $el.data('paginator', pub);
        return pub;
    }

    $.fn.paginate = function paginate(opts) {
        return new Paginator(this, opts);
    };
})();
